import os
import json
import sqlparse
import time
import glob

import networkx as nx

from copy import deepcopy
from networkx.readwrite import json_graph
from pathlib import Path
from pyhocon import ConfigFactory
from typing import Dict, List

from meltano.core.plugin import PluginType
from meltano.core.plugin_discovery_service import PluginDiscoveryService
from meltano.core.sql.design_helper import PypikaJoinExecutor
from meltano.core.project import Project
from meltano.core.plugin.model import Package
from meltano.core.utils import encode_id_from_file_path, slugify, find_named


class MeltanoAnalysisFileParserError(Exception):
    def __init__(self, message, file_name):
        self.message = message
        self.file_name = file_name
        super().__init__(f"{file_name}: {message}")


class MeltanoAnalysisFileParserMissingTableError(MeltanoAnalysisFileParserError):
    def __init__(self, field, your_choice, cls, file_name):
        message = (
            f'Missing accompanying table "{your_choice}" in "{field}" field in {cls}'
        )
        super().__init__(message, file_name)


class MeltanoAnalysisMissingTopicFilesError(MeltanoAnalysisFileParserError):
    def __init__(self, message):
        super().__init__(message, "topic file")


class MeltanoAnalysisFileParserUnacceptableChoiceError(MeltanoAnalysisFileParserError):
    def __init__(self, field, your_choice, acceptable_choices, cls, file_name):
        quoted_choices = [f'"{a}"' for a in acceptable_choices]
        message = f"Unacceptable choice for field {field}. You wrote \"{your_choice}\". Acceptable choices are: {', '.join(quoted_choices)}"
        super().__init__(message, file_name)


class MeltanoAnalysisFileParserMissingFieldsError(MeltanoAnalysisFileParserError):
    def __init__(self, fields, cls, file_name):
        field = "field" if len(fields) == 1 else "fields"
        message = f"Missing {field} in {cls}: \"{', '.join(fields)}\""
        super().__init__(message, file_name)


class MeltanoAnalysisInvalidJoinDependencyError(MeltanoAnalysisFileParserError):
    def __init__(self, field, cls, file_name):
        message = f'Invalid join dependency for join {cls} in in field "{field}"'
        super().__init__(message, file_name)


class MeltanoAnalysisFileParser:
    def __init__(self, project: Project):
        self.project = project
        self._output_dir = self.project.run_dir("models")
        self.topics = []
        self.tables = []
        self.packaged_topics = []
        self.packaged_tables = []
        self.required_topic_properties = ["name", "connection", "label", "designs"]
        self.required_design_properties = ["from", "label", "description"]
        self.required_join_properties = ["sql_on", "relationship"]
        self.required_table_properties = ["sql_table_name", "columns"]
        self.join_relationship_types = [
            "one_to_one",
            "one_to_many",
            "many_to_one",
            "many_to_many",
        ]
        self.project = Project()

    def uses_accepted_choice(self, choices, choice):
        return choice in choices

    def missing_properties(self, properties, properties_dict):
        properties_copy = properties.copy()
        for prop in properties_dict:
            try:
                property_index = properties_copy.index(prop)
                del properties_copy[property_index]
            except ValueError as e:
                continue
        return properties_copy

    def parse_m5o_file(self, file_path):
        try:
            parsed = ConfigFactory.parse_file(file_path)
            parsed["_file_path"] = str(file_path)
            return parsed
        except Exception as e:
            raise MeltanoAnalysisFileParserError(str(e), str(file_path.parts[-1]))

    def generate_join_graph_for_node(self, graph, design, source_name, joins):
        remaining_joins = []
        check_against = []

        for join in joins:
            join_executor = None
            if "executor" in join:
                join_executor = join["executor"]
            else:
                join_executor = join["executor"] = PypikaJoinExecutor(design, join)
                join_executor.visit(sqlparse.parse(join["sql_on"])[0])

            # the `table` identifier here represent either:
            #   - Design name
            #   - Join name
            left_join_candidate = join_executor.comparison_fields.left.table
            right_join_candidate = join_executor.comparison_fields.right.table

            if left_join_candidate == source_name:
                check_against.append(right_join_candidate)

                # that means we must add the `right` join
                graph.add_node(right_join_candidate)
                graph.add_edge(left_join_candidate, right_join_candidate)
            elif right_join_candidate == source_name:
                check_against.append(left_join_candidate)

                # that means we must add the `left` join
                graph.add_node(left_join_candidate)
                graph.add_edge(right_join_candidate, left_join_candidate)
            else:
                # both `left` & `right` are joins
                remaining_joins.append(join)

        # recurse through the remaining joins
        for check in check_against:
            self.generate_join_graph_for_node(graph, design, check, remaining_joins)

    def graph_design(self, design):
        # The `base` table should always be part of the graph
        # base_table = design["related_table"]["name"]
        source_name = design["name"]

        graph = nx.DiGraph()
        graph.add_node(source_name)
        joins = deepcopy(design.get("joins", []))

        self.generate_join_graph_for_node(graph, design, source_name, joins)
        return json_graph.node_link_data(graph)

    def graph_topic(self, topic):
        for design in topic["designs"]:
            design_graph = self.graph_design(design)
            design["graph"] = design_graph

        return topic

    def compile(self, topics):
        indices = {}
        for topic in topics:
            # Each topic is:
            # - indexed using its topic['namespace']
            #   e.g. package.name for a package
            #   `cutom/dir1/dir2` for a custom topic defined under `project/models/dir1/dir2/`
            # - compiled and stored under `.meltano/run/models/{topic['namespace']}`
            #   e.g. `.meltano/run/models/custom/dir1/dir2/mytopic.topic.m5oc`
            topic_identifier = Path(topic["namespace"]).joinpath(topic["name"])

            topic = self.graph_topic(topic)
            indices[str(topic_identifier)] = {
                "plugin_namespace": topic["plugin_namespace"],
                "namespace": topic["namespace"],
                "name": topic["name"],
                "designs": [e["name"] for e in topic["designs"]],
            }

            # Recursively create the missing directories under `.meltano/run/`
            compiled_file_dir = self._output_dir.joinpath(topic["namespace"])
            compiled_file_dir.mkdir(parents=True, exist_ok=True)

            # Create the m5oc file for the topic
            compiled_file_name = f"{topic['name']}.topic.m5oc"
            compiled_file_path = compiled_file_dir.joinpath(compiled_file_name)
            with compiled_file_path.open("w") as compiled_topic:
                json.dump(topic, compiled_topic)

        # Add the final index file under `.meltano/run/models/`
        index_file_path = self._output_dir.joinpath("topics.index.m5oc")
        with index_file_path.open("w") as index_file:
            index_file.write(json.dumps(indices))

    def parse_packages(self):

        discovery = PluginDiscoveryService(self.project)
        model_plugins = [
            plugin for plugin in discovery.plugins() if plugin.type is PluginType.MODELS
        ]

        for package in self.packages():
            if not package.topics and package.tables:
                raise MeltanoAnalysisMissingTopicFilesError(
                    f"Missing topic file(s) for package {package}"
                )

            for table in package.tables:
                conf = self.parse_m5o_file(table)
                parsed_table = self.table(conf, table.name)
                self.tables.append(parsed_table)

            for topic in package.topics:
                conf = self.parse_m5o_file(topic)
                parsed_topic = self.topic(conf, topic.name)
                model = next(
                    (plugin for plugin in model_plugins if plugin.name == package.name),
                    None,
                )
                parsed_topic["plugin_namespace"] = model.namespace if model else None
                parsed_topic["namespace"] = package.name
                self.packaged_topics.append(parsed_topic)

            # Reset the tables list so that tables with the same name from
            #  different packages are not interfearing with correctly parsing
            #  the packages that follow
            self.tables = []

        return self.packaged_topics

    def parse(self):
        # Iterate through the project/models/ directory and all its subdirectories
        # Each directory sets a namespace for the topics and tables that exist
        #  inside it and all the tables and topics in each namespace are parsed
        #  together (so we assume unique names and they can reference each other).
        models = self.project.root_dir("model")
        subfolders = [f for f in models.glob("**/*") if f.is_dir()]

        for folder in [models] + subfolders:
            self.m5o_tables = list(folder.glob("*.table.m5o"))
            self.m5o_topics = list(folder.glob("*.topic.m5o"))
            if not self.m5o_topics and self.m5o_tables:
                raise MeltanoAnalysisMissingTopicFilesError("Missing topic file(s)")

            for table in self.m5o_tables:
                conf = self.parse_m5o_file(table)
                parsed_table = self.table(conf, table.name)
                self.tables.append(parsed_table)

            for topic in self.m5o_topics:
                conf = self.parse_m5o_file(topic)
                parsed_topic = self.topic(conf, topic.name)

                # The namespace for a topic in a directory under project/models/
                #   is `custom/path/to/directory`
                # For example:
                # `project/models/mytopic.topic.m5o` --> `cutom`
                # `project/models/dir1/dir2/mytopic.topic.m5o` --> `cutom/dir1/dir2`
                parsed_topic["namespace"] = str(
                    Path("custom").joinpath(folder.relative_to(models))
                )

                parsed_topic["plugin_namespace"] = parsed_topic["namespace"].replace(
                    "/", "_"
                )

                self.topics.append(parsed_topic)

            self.tables = []

        return self.topics

    def topic(self, ma_file_topic_dict, file_name):
        temp_topic = {}
        missing_properties = self.missing_properties(
            self.required_topic_properties, ma_file_topic_dict
        )

        if missing_properties:
            raise MeltanoAnalysisFileParserMissingFieldsError(
                missing_properties, "topic", file_name
            )

        for prop_name, prop_def in ma_file_topic_dict.items():
            temp_topic[prop_name] = prop_def
            if prop_name == "designs":
                temp_topic[prop_name] = self.designs(prop_def, file_name)

        return temp_topic

    def table_conf_by_name(self, table_name, cls, prop, file_name):
        try:
            return find_named(self.tables, table_name)
        except StopIteration as e:
            raise MeltanoAnalysisFileParserMissingTableError(
                prop, table_name, cls, file_name
            )

    def designs(self, ma_file_designs_dict, file_name):
        topic_designs = []
        for design_name, design_def in ma_file_designs_dict.items():
            temp_design = {}
            temp_design["name"] = design_name
            missing_properties = self.missing_properties(
                self.required_design_properties, design_def
            )
            if missing_properties:
                raise MeltanoAnalysisFileParserMissingFieldsError(
                    missing_properties, "design", file_name
                )

            for prop_name, prop_def in design_def.items():
                temp_design[prop_name] = prop_def
                if prop_name == "from":
                    temp_design["related_table"] = self.table_conf_by_name(
                        temp_design[prop_name], "design", prop_name, file_name
                    )
                if prop_name == "joins":
                    temp_design[prop_name] = self.joins(prop_def, file_name)

            topic_designs.append(temp_design)

        return topic_designs

    def joins(self, ma_file_joins_dict, file_name):
        design_joins = []
        for join_name, join_def in ma_file_joins_dict.items():
            temp_join = {}
            temp_join["name"] = join_name
            related_table_name = join_def.get("from", join_name)
            temp_join["related_table"] = self.table_conf_by_name(
                related_table_name, "join", "name", file_name
            )
            missing_properties = self.missing_properties(
                self.required_join_properties, join_def
            )
            if missing_properties:
                raise MeltanoAnalysisFileParserMissingFieldsError(
                    missing_properties, "join", file_name
                )

            for prop_name, prop_def in join_def.items():
                temp_join[prop_name] = prop_def
                if prop_name == "relationship":
                    uses_accepted_choices = self.uses_accepted_choice(
                        self.join_relationship_types, prop_def
                    )
                    if not uses_accepted_choices:
                        raise MeltanoAnalysisFileParserUnacceptableChoiceError(
                            prop_name,
                            prop_def,
                            self.join_relationship_types,
                            "join",
                            file_name,
                        )

            design_joins.append(temp_join)

        return design_joins

    def table(self, table_file, file_name):
        temp_table = {}
        missing_properties = self.missing_properties(
            self.required_table_properties, table_file
        )
        if missing_properties:
            raise MeltanoAnalysisFileParserMissingFieldsError(
                missing_properties, "table", file_name
            )

        for prop_name, prop_def in table_file.items():
            subproperties = ("columns", "aggregates", "timeframes")

            if prop_name in subproperties:
                temp_table[prop_name] = self.name_flatten_dict(prop_def)
            else:
                temp_table[prop_name] = prop_def

        return temp_table

    @classmethod
    def name_flatten_dict(cls, d: Dict) -> List:
        return [{"name": k, **rest} for k, rest in d.items()]

    @staticmethod
    def fill_base_m5o_dict(file, name, file_dict=None):
        if file_dict is None:
            file_dict = {"name": name}

        file_dict["path"] = str(file)
        file_dict["id"] = encode_id_from_file_path(file_dict["path"])
        file_dict["slug"] = slugify(name)
        file_dict["createdAt"] = time.time()
        return file_dict

    def packages(self) -> List[Package]:
        return [
            Package(f.name, self.project)
            for f in self.project.model_dir().iterdir()
            if f.is_dir()
        ]

    @classmethod
    def package_files(cls, package):
        m5oFiles = {
            "topics": {"label": "Topics", "items": []},
            "tables": {"label": "Tables", "items": []},
        }
        for f in package.files:
            basename = os.path.basename(f)
            filename, ext = os.path.splitext(basename)
            if ext != ".m5o":
                continue

            # filename splittext twice occurs due to current *.type.extension convention (two dots)
            filename = f"{package.name}/{filename.lower()}"
            filename, ext = os.path.splitext(filename)
            file_dict = cls.fill_base_m5o_dict(f, filename)
            if ext == ".topic":
                m5oFiles["topics"]["items"].append(file_dict)
            if ext == ".table":
                m5oFiles["tables"]["items"].append(file_dict)

        return m5oFiles



import functools
import urllib
import click
import click.globals
import os
from pathlib import Path

from meltano.core.project import Project
from meltano.core.utils import pop_all
from meltano.core.migration_service import MigrationService
from meltano.core.db import project_engine


def db_options(func):
    @click.option(
        "--database-uri",
        envvar="MELTANO_DATABASE_URI",
        default=lambda: f"sqlite:///{os.getcwd()}/.meltano/meltano.db",
        help="System database URI",
    )
    def decorate(*args, **kwargs):
        engine_uri = kwargs.pop("database_uri")

        return func(engine_uri, *args, **kwargs)

    return functools.update_wrapper(decorate, func)


class project:
    __name__ = "project"

    def __init__(self, migrate=False):
        self.migrate = migrate

    def __call__(self, func):
        @db_options
        def decorate(engine_uri, *args, **kwargs):
            ctx = click.globals.get_current_context()

            project = ctx.obj["project"]
            if not project:
                raise click.ClickException(
                    f"`{ctx.command_path}` must be run inside a Meltano project."
                    "\nUse `meltano init <project_name>` to create one."
                )

            # register the system database connection
            engine, _ = project_engine(project, engine_uri, default=True)

            if self.migrate:
                migration_service = MigrationService(engine)
                migration_service.upgrade()
                migration_service.seed(project)

            func(project, *args, **kwargs)

        return functools.update_wrapper(decorate, func)



from enum import Enum
from pypika import functions as fn

from .substitution import Substitution


class AggregateType(Enum):
    Unknown = "UNKNOWN"
    Count = "count"
    Sum = "sum"
    Avg = "avg"
    Number = "number"

    def __eq__(self, other):
        return self.value.lower() == other.value.lower()

    @classmethod
    def parse(cls, value: str):
        try:
            return next(e for e in cls if e.value.lower() == value)
        except StopIteration:
            return cls.Unknown


class Aggregate:
    def __init__(self, aggregate, db_table):
        self.aggregate = aggregate
        self.field = Substitution(
            aggregate["sql"], db_table, None, alias=aggregate["name"]
        )
        self.aggregate_type = AggregateType.parse(aggregate["type"])

    @property
    def alias(self):
        return f"{self.field.table.alias}.{self.aggregate['name']}"

    @property
    def sql(self):
        if self.aggregate_type == AggregateType.Unknown:
            raise NotImplementedError(f"Unknown aggregate type.")

        return getattr(self, self.aggregate_type.value)()

    def sum(self):
        return fn.Coalesce(fn.Sum(self.field.sql), 0, alias=self.alias)

    def count(self):
        return fn.Coalesce(fn.Count(self.field.sql), 0, alias=self.alias)

    def avg(self):
        return fn.Coalesce(fn.Avg(self.field.sql), 0, alias=self.alias)

    def number(self):
        self.field.alias = self.field.alias
        return self.field.sql



import json
import os

from os.path import join
from pathlib import Path

from meltano.core.m5o.m5oc_file import M5ocFile
from meltano.core.m5o.m5o_file_parser import MeltanoAnalysisFileParser
from meltano.core.m5o.m5o_collection_parser import (
    M5oCollectionParser,
    M5oCollectionParserTypes,
)
from meltano.core.project import Project
from meltano.core.utils import slugify
from meltano.core.m5o.m5o_collection_parser import (
    M5oCollectionParser,
    M5oCollectionParserTypes,
)


class ReportAlreadyExistsError(Exception):
    """Occurs when a report already exists."""

    def __init__(self, report):
        self.report = report


class ReportsHelper:
    VERSION = "1.0.0"

    def get_report_by_name(self, name):
        reports = self.get_reports()
        report = next(filter(lambda r: r["name"] == name, reports), None)
        return report

    def get_reports(self):
        project = Project.find()
        reportsParser = M5oCollectionParser(
            project.analyze_dir("reports"), M5oCollectionParserTypes.Report
        )

        return reportsParser.parse()

    def load_report(self, report_name):
        return self.get_report_by_name(report_name)

    def save_report(self, data):
        report_name = data["name"]

        # guard if it already exists
        existing_report = self.get_report_by_name(report_name)
        if existing_report:
            raise ReportAlreadyExistsError(existing_report)

        project = Project.find()
        slug = slugify(report_name)
        file_path = project.analyze_dir("reports", f"{slug}.report.m5o")

        data = MeltanoAnalysisFileParser.fill_base_m5o_dict(file_path, slug, data)
        data["version"] = ReportsHelper.VERSION

        with file_path.open("w") as f:
            json.dump(data, f)

        return data

    def update_report(self, data):
        project = Project.find()
        file_path = project.analyze_dir("reports", f"{data['slug']}.report.m5o")
        with open(file_path, "w") as f:
            json.dump(data, f)

        return data



import os
import json
import click
from urllib.parse import urlparse

from . import cli
from .params import project
from meltano.core.plugin import PluginType
from meltano.core.project import Project
from meltano.core.plugin_discovery_service import (
    PluginDiscoveryService,
    DiscoveryInvalidError,
)
from meltano.core.tracking import GoogleAnalyticsTracker


@cli.command()
@click.argument(
    "plugin_type",
    type=click.Choice(
        [
            PluginType.EXTRACTORS,
            PluginType.LOADERS,
            PluginType.TRANSFORMERS,
            PluginType.MODELS,
            PluginType.TRANSFORMS,
            PluginType.ORCHESTRATORS,
            PluginType.ALL,
        ]
    ),
)
@project()
def discover(project, plugin_type):
    discover_service = PluginDiscoveryService(project)
    try:
        discovery_dict = discover_service.discover(plugin_type)

        for plugin_type, plugins in discovery_dict.items():
            click.secho(plugin_type, fg="green")
            for plugin in plugins:
                click.echo(plugin)

        tracker = GoogleAnalyticsTracker(project)
        tracker.track_meltano_discover(plugin_type=plugin_type)
    except Exception as e:
        click.secho("Cannot list available plugins.", fg="red")
        raise click.ClickException(str(e))



import cerberus
import logging
import yaml

from typing import Dict, List
from meltano.core.permissions.utils.error import SpecLoadingError
from meltano.core.permissions.spec_schemas.postgres import *


VALIDATION_ERR_MSG = 'Spec error: Role "{}", field "{}": {}'


class PGSpecLoader:
    def __init__(self, spec_path: str) -> None:
        self.spec = self.load_spec(spec_path)

    def load_spec(self, spec_path: str) -> Dict:
        """
        Load a permissions specification from a file.

        If the file is not found or at least an error is found during validation,
        raise a SpecLoadingError with the appropriate error messages.

        Otherwise, return the valid specification as a Dictionary to be used
        in other operations.
        """
        try:
            with open(spec_path, "r") as stream:
                spec = yaml.load(stream)
        except FileNotFoundError:
            raise SpecLoadingError(f"Spec File {spec_path} not found")

        error_messages = self.ensure_valid_schema(spec)
        if error_messages:
            raise SpecLoadingError("\n".join(error_messages))

        return spec

    def ensure_valid_schema(self, spec: Dict) -> List[str]:
        """
        Ensure that the provided spec has no schema errors.

        Return a list with all the errors found.
        """
        error_messages = []

        schema = yaml.load(PG_SPEC_SCHEMA_YAML)
        v = cerberus.Validator(schema)

        role_configs = [(role, config) for role, config in spec.items() if config]

        for role, config in role_configs:
            v.validate(config)
            for field, err_msg in v.errors.items():
                error_messages.append(
                    VALIDATION_ERR_MSG.format(role, field, err_msg[0])
                )

        return error_messages

    def generate_permission_queries(self) -> List[str]:
        sql_commands = []

        role_configs = [(role, config) for role, config in self.spec.items() if config]

        for role, config in role_configs:
            sql_commands.extend(self.generate_alter_role(role, config))
            sql_commands.extend(self.generate_grant_roles_to_role(role, config))
            sql_commands.extend(self.generate_grant_ownership_to_role(role, config))
            sql_commands.extend(self.generate_grant_privileges_to_role(role, config))

        return sql_commands

    def generate_alter_role(self, role: str, config: Dict) -> List[str]:
        ALTER_ROLE_TEMPLATE = "ALTER ROLE {role} {privileges}"
        sql_commands = []
        alter_privileges = []

        try:
            if config["can_login"]:
                alter_privileges.append("LOGIN")
            else:
                alter_privileges.append("NOLOGIN")
        except KeyError:
            logging.debug(
                "`can_login` not found for {}, skipping login rules.".format(role)
            )

        try:
            if config["is_superuser"]:
                alter_privileges.append("SUPERUSER")
            else:
                alter_privileges.append("NOSUPERUSER")
        except KeyError:
            logging.debug(
                "`is_superuser` not found for {}, skipping superuser rules.".format(
                    role
                )
            )

        if alter_privileges:
            sql_commands.append(
                {
                    "already_granted": False,
                    "sql": ALTER_ROLE_TEMPLATE.format(
                        role=role, privileges=" ".join(alter_privileges)
                    ),
                }
            )

        return sql_commands

    def generate_grant_roles_to_role(self, role: str, config: Dict) -> List[str]:
        GRANT_ROLE_TEMPLATE = "GRANT {role_names} TO {role}"
        sql_commands = []

        role_names = config.get("member_of", [])

        if role_names:
            sql_commands.append(
                {
                    "already_granted": False,
                    "sql": GRANT_ROLE_TEMPLATE.format(
                        role=role, role_names=", ".join(role_names)
                    ),
                }
            )

        return sql_commands

    def generate_grant_ownership_to_role(self, role: str, config: Dict) -> List[str]:
        ALTER_SCHEMA_OWNER_TEMPLATE = "ALTER SCHEMA {schema} OWNER TO {role}"
        sql_commands = []

        try:
            for schema in config["owns"]["schemas"]:
                sql_commands.append(
                    {
                        "already_granted": False,
                        "sql": ALTER_SCHEMA_OWNER_TEMPLATE.format(
                            role=role, schema=schema
                        ),
                    }
                )
        except KeyError:
            logging.debug(
                "`owns.schemas` not found for {}, skipping OWNership rules.".format(
                    role
                )
            )

        return sql_commands

    def generate_grant_privileges_to_role(self, role: str, config: Dict) -> List[str]:
        GRANT_READ_ON_SCHEMA_TEMPLATE = "GRANT USAGE ON SCHEMA {schema} TO {role}"
        GRANT_WRITE_ON_SCHEMA_TEMPLATE = "GRANT CREATE ON SCHEMA {schema} TO {role}"
        GRANT_READ_ON_TABLE_TEMPLATE = "GRANT SELECT ON TABLE {table} TO {role}"
        GRANT_WRITE_ON_TABLE_TEMPLATE = (
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES, TRIGGER "
            "ON TABLE {table} TO {role}"
        )
        GRANT_READ_ON_ALL_TABLES_TEMPLATE = (
            "GRANT SELECT ON ALL TABLES IN SCHEMA {schema} TO {role}"
        )
        GRANT_WRITE_ON_ALL_TABLE_TEMPLATE = (
            "GRANT SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES, TRIGGER "
            "ON ALL TABLES IN SCHEMA {schema} TO {role}"
        )

        sql_commands = []

        try:
            for schema in config["privileges"]["schemas"]["read"]:
                sql_commands.append(
                    {
                        "already_granted": False,
                        "sql": GRANT_READ_ON_SCHEMA_TEMPLATE.format(
                            role=role, schema=schema
                        ),
                    }
                )
        except KeyError:
            logging.debug(
                "`privileges.schemas.read` not found for {}, skipping Schema Read GRANTS.".format(
                    role
                )
            )

        try:
            for schema in config["privileges"]["schemas"]["write"]:
                sql_commands.append(
                    {
                        "already_granted": False,
                        "sql": GRANT_WRITE_ON_SCHEMA_TEMPLATE.format(
                            role=role, schema=schema
                        ),
                    }
                )
        except KeyError:
            logging.debug(
                "`privileges.schemas.write` not found for {}, skipping Schema Write GRANTS.".format(
                    role
                )
            )

        try:
            for table in config["privileges"]["tables"]["read"]:
                if table.endswith(".*"):
                    schema = table[:-2]
                    sql_commands.append(
                        {
                            "already_granted": False,
                            "sql": GRANT_READ_ON_ALL_TABLES_TEMPLATE.format(
                                role=role, schema=schema
                            ),
                        }
                    )
                else:
                    sql_commands.append(
                        {
                            "already_granted": False,
                            "sql": GRANT_READ_ON_TABLE_TEMPLATE.format(
                                role=role, table=table
                            ),
                        }
                    )
        except KeyError:
            logging.debug(
                "`privileges.tables.read` not found for {}, skipping Table Read GRANTS.".format(
                    role
                )
            )

        try:
            for table in config["privileges"]["tables"]["write"]:
                if table.endswith(".*"):
                    schema = table[:-2]
                    sql_commands.append(
                        {
                            "already_granted": False,
                            "sql": GRANT_WRITE_ON_ALL_TABLE_TEMPLATE.format(
                                role=role, schema=schema
                            ),
                        }
                    )
                else:
                    sql_commands.append(
                        {
                            "already_granted": False,
                            "sql": GRANT_WRITE_ON_TABLE_TEMPLATE.format(
                                role=role, table=table
                            ),
                        }
                    )
        except KeyError:
            logging.debug(
                "`privileges.tables.write` not found for {}, skipping Table Write GRANTS.".format(
                    role
                )
            )

        return sql_commands



import yaml
import fnmatch
from typing import Dict, Union
from collections import namedtuple
from enum import Enum
from typing import Optional

from meltano.core.behavior.hookable import HookObject
from meltano.core.utils import compact


class YAMLEnum(str, Enum):
    def __str__(self):
        return self.value

    @staticmethod
    def yaml_representer(dumper, obj):
        return dumper.represent_scalar("tag:yaml.org,2002:str", str(obj))


yaml.add_multi_representer(YAMLEnum, YAMLEnum.yaml_representer)


class PluginType(YAMLEnum):
    EXTRACTORS = "extractors"
    LOADERS = "loaders"
    MODELS = "models"
    TRANSFORMERS = "transformers"
    TRANSFORMS = "transforms"
    ORCHESTRATORS = "orchestrators"
    ALL = "all"

    def __str__(self):
        return self.value

    @property
    def cli_command(self):
        """Makes it singular for `meltano add PLUGIN_TYPE`"""
        if self is self.__class__.ALL:
            raise NotImplemented()

        return self.value[:-1]

    @classmethod
    def value_exists(cls, value):
        return value in cls._value2member_map_


class PluginRef:
    def __init__(self, plugin_type: Union[str, PluginType], name: str):
        self.type = (
            plugin_type
            if isinstance(plugin_type, PluginType)
            else PluginType(plugin_type)
        )
        self.canonical_name, self.profile = self.parse_name(name)

    @classmethod
    def parse_name(cls, name: str):
        name, *profile = name.split("@")
        profile = next(iter(profile), None)

        return (name, profile)

    @property
    def qualified_name(self):
        parts = (self.type, self.canonical_name, self.profile)

        return ".".join(compact(parts))

    @property
    def name(self):
        parts = (self.canonical_name, self.profile)

        return "@".join(compact(parts))

    def __eq__(self, other):
        return self.name == other.name and self.type == other.type

    def __hash__(self):
        return hash((self.type, self.canonical_name, self.profile))


class PluginInstall(HookObject, PluginRef):
    def __init__(
        self,
        plugin_type: PluginType,
        name: str,
        pip_url: Optional[str] = None,
        select=set(),
        config={},
        **extras
    ):
        super().__init__(plugin_type, name)

        self.pip_url = pip_url
        self._config = config
        self.select = set(select)
        self._extras = extras or {}

    def canonical(self):
        canonical = {"name": self.name, **self._extras}

        if self.pip_url:
            canonical.update({"pip_url": self.pip_url})

        if self.select:
            canonical.update({"select": list(self.select)})

        if self._config:
            canonical.update({"config": self._config})

        return canonical

    def installable(self):
        return self.pip_url is not None

    @property
    def config(self):
        if not self.profile:
            return self._config

        return self._extras["profiles"][self.profile]

    @property
    def executable(self):
        return self._extras.get("executable", self.canonical_name)

    def exec_args(self, files: Dict):
        return []

    @property
    def config_files(self):
        """Return a list of stubbed files created for this plugin."""
        return dict()

    @property
    def output_files(self):
        return dict()

    def add_select_filter(self, filter: str):
        self.select.add(filter)


class Plugin(PluginRef):
    """
    Args:
    name: The unique name for the installed plugin
    pip_url: The pip-compatible installation URI, like `git+https://…` or `-e /path/to/pkg`
    executable: The plugin executable name (default: <name>)
    """

    def __init__(
        self,
        plugin_type: PluginType,
        name: str,
        namespace: str,
        pip_url: Optional[str] = None,
        settings: list = [],
        docs=None,
        description=None,
        capabilities=set(),
        select=set(),
        **extras
    ):
        super().__init__(plugin_type, name)

        self.namespace = namespace
        self.pip_url = pip_url
        self.settings = settings
        self.docs = docs
        self.description = description
        self.capabilities = set(capabilities)
        self.select = set(select)
        self._extras = extras or {}

    def canonical(self):
        canonical = {"name": self.name, "namespace": self.namespace, **self._extras}

        if self.pip_url:
            canonical.update({"pip_url": self.pip_url})

        if self.docs:
            canonical.update({"docs": self.docs})

        if self.description:
            canonical.update({"description": self.description})

        if self.settings:
            canonical.update({"settings": self.settings})

        if self.capabilities:
            canonical.update({"capabilities": list(self.capabilities)})

        if self.select:
            canonical.update({"select": list(self.select)})

        return canonical

    def as_installed(self) -> PluginInstall:
        return PluginInstall(
            self.type,
            self.name,
            pip_url=self.pip_url,
            select=self.select,
            **self._extras
        )



#!/usr/bin/env python
from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("VERSION") as version_file:
    version = version_file.read().strip()

requires = [
    'aiohttp==3.4.4',
    'alembic==1.0.11',
    'atomicwrites==1.2.1',
    'authlib==0.10',
    'backoff==1.8.0',
    'bcrypt==3.1.6',
    'Cerberus==1.2',
    'click==7.0',
    'click-default-group==1.2.1',
    'colorama==0.3.9',
    'gitpython==2.1.11',
    'gunicorn==19.9.0',
    'jsonschema==2.6.0',
    'markdown==3.0.1',
    'networkx==2.2',
    'pandas==0.24.1',
    'psycopg2==2.7.7',
    'psutil==5.6.3',
    'pyhumps==0.4.0',
    'pypika==0.25.1',
    'python-dotenv==0.10.1',
    'pyyaml==3.13',
    'snowflake-connector-python==1.6.10',
    'snowflake-sqlalchemy==1.1.2',
    'sqlalchemy==1.2.12',
    'fasteners==0.15.0',
    'flask>=1.0.2',
    'flask-cors==3.0.7',
    'flask-executor==0.9.2',
    'flask-sqlalchemy==2.3.2',
    'flask-restful==0.3.7',
    'flask-jwt-extended==3.17.0',
    'flatten-dict==0.1.0',
    'meltano-flask-security==0.1.0',
    'pyhocon==0.3.51',
    'python-dotenv==0.10.1',
    'python-gitlab==1.8.0',
    'simplejson==3.16.0',
    'urllib3==1.23',
    'sqlparse==0.3.0',
    'watchdog==0.9.0',
]

# conflicts resolution, see https://gitlab.com/meltano/meltano/issues/193
conflicts = [
    'aenum==2.1.2',
    'idna==2.7',
    'asn1crypto==0.24.0',
]

dev_requires = [
    'asynctest==0.12.2',
    'black==18.9b0',
    'bumpversion==0.5.3',
    'changelog-cli==0.6.2',
    'freezegun==0.3.12',
    'pytest==4.3.1',
    'pytest-asyncio==0.10.0',
    'requests-mock==1.6.0',
    'ipython==7.5.0',
    'pytest-cov==2.6.1',
]

setup(
    name="meltano",
    version=version,
    author='Meltano Team & Contributors',
    author_email="meltano@gitlab.com",
    description="Meltano",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/meltano/meltano",
    package_dir={'': 'src'},
    packages=find_packages(where="src"),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    setup_requires=['pytest-runner'],
    tests_require=dev_requires,
    # run `make requirements.txt` after editing
    install_requires=[
        *conflicts,
        *requires
    ],
    extras_require={
        'dev': dev_requires
    },
    entry_points={
        'console_scripts': [
            "meltano = meltano.cli:main"
        ]
    }
)



from pyhocon import ConfigFactory
from pathlib import Path
from enum import Enum

import json


class M5oCollectionParserError(Exception):
    def __init__(self, message, file_name, *args):
        self.message = message
        self.file_name = file_name
        super(M5oCollectionParserError, self).__init__(
            self.message, self.file_name, *args
        )


class M5oCollectionParserTypes(Enum):
    Dashboard = "dashboard"
    Report = "report"


class M5oCollectionParser:
    def __init__(self, directory, file_type):
        self.directory = directory
        self.file_type = file_type
        self.pattern = f"*.{self.file_type.value}.m5o"
        self.files = []

    def parse(self):
        files = []
        for file in Path(self.directory).glob(self.pattern):
            file_name = file.parts[-1]
            m5o_file = Path(self.directory).joinpath(file_name)
            with m5o_file.open() as f:
                files.append(json.load(f))

        return files



import os
import subprocess
import click
import signal
import logging
import asyncio

from . import cli
from .params import project
from meltano.core.db import project_engine
from meltano.core.tracking import GoogleAnalyticsTracker
from meltano.core.utils import truthy
from meltano.core.migration_service import MigrationService
from meltano.api.workers import (
    MeltanoCompilerWorker,
    AirflowWorker,
    APIWorker,
    UIAvailableWorker,
    DbtWorker,
)


logger = logging.getLogger(__name__)


def start_workers(workers):
    def stop_all():
        logger.info("Stopping all background workers...")
        for worker in workers:
            worker.stop()

    # start all workers
    for worker in workers:
        worker.start()

    return stop_all


@cli.command()
@click.option("--reload", is_flag=True, default=False)
@click.option(
    "--bind-port",
    default=5000,
    help="Port to run webserver on",
    envvar="MELTANO_API_PORT",
    type=int,
)
@click.option(
    "--bind",
    default="0.0.0.0",
    help="The hostname (or IP address) to bind on",
    envvar="MELTANO_API_HOSTNAME",
)
@project(migrate=True)
def ui(project, reload, bind_port, bind):
    tracker = GoogleAnalyticsTracker(project)
    tracker.track_meltano_ui()

    loop = asyncio.get_event_loop()

    # we need to prime the ChildWatcher here so we can
    # call subprocesses asynchronously from threads
    #
    # see https://docs.python.org/3/library/asyncio-subprocess.html#subprocess-and-threads
    # TODO: remove when running on Python 3.8
    asyncio.get_child_watcher()

    workers = []
    if not truthy(os.getenv("MELTANO_DISABLE_AIRFLOW", False)):
        workers.append(AirflowWorker(project))

    workers.append(MeltanoCompilerWorker(project))

    # we need to whitelist the loaders here because not
    # all the loaders support dbt in the first place
    dbt_docs_loader = os.getenv("MELTANO_DBT_DOCS_LOADER", "target-postgres")
    if dbt_docs_loader:
        workers.append(DbtWorker(project, dbt_docs_loader, loop=loop))
    else:
        logging.info(
            "No loader enabled for dbt docs generation, set the MELTANO_DBT_DOCS_LOADER variable to enable one."
        )

    workers.append(UIAvailableWorker("http://localhost:{bind_port}"))
    workers.append(
        APIWorker(
            project,
            f"{bind}:{bind_port}",
            reload=reload or os.getenv("FLASK_ENV") == "development",
        )
    )

    cleanup = start_workers(workers)

    def handle_terminate(signal, frame):
        cleanup()

    signal.signal(signal.SIGTERM, handle_terminate)
    logger.info("All workers started.")



import logging

from flask import Blueprint, request, url_for, jsonify, make_response, Response

from meltano.core.job import JobFinder, State
from meltano.core.plugin import PluginRef
from meltano.core.plugin.error import PluginExecutionError, PluginLacksCapabilityError
from meltano.core.plugin.settings_service import (
    PluginSettingsService,
    PluginSettingValueSource,
)
from meltano.core.plugin_discovery_service import PluginDiscoveryService
from meltano.core.plugin_install_service import PluginInstallService
from meltano.core.project import Project
from meltano.core.project_add_service import ProjectAddService
from meltano.core.config_service import ConfigService
from meltano.core.schedule_service import ScheduleService, ScheduleAlreadyExistsError
from meltano.core.utils import flatten, iso8601_datetime, slugify
from meltano.core.logging import JobLoggingService, MissingJobLogException
from meltano.cli.add import extractor
from meltano.api.models import db
from meltano.api.json import freeze_keys

from meltano.api.executor import run_elt


orchestrationsBP = Blueprint(
    "orchestrations", __name__, url_prefix="/api/v1/orchestrations"
)


@orchestrationsBP.errorhandler(ScheduleAlreadyExistsError)
def _handle(ex):
    return (
        jsonify(
            {
                "error": True,
                "code": f"A schedule with the name '{ex.schedule.name}' already exists. Try renaming the schedule.",
            }
        ),
        409,
    )


@orchestrationsBP.errorhandler(MissingJobLogException)
def _handle(ex):
    return (jsonify({"error": False, "code": str(ex)}), 204)


@orchestrationsBP.route("/jobs/state", methods=["POST"])
def job_state() -> Response:
    """
    Endpoint for getting the status of N jobs
    """
    project = Project.find()
    poll_payload = request.get_json()
    job_ids = poll_payload["job_ids"]

    jobs = []
    for job_id in job_ids:
        finder = JobFinder(job_id)
        state_job = finder.latest(db.session)
        # Validate existence first as a job may not be queued yet as a result of
        # another prerequisite async process (dbt installation for example)
        if state_job:
            jobs.append(
                {
                    "job_id": job_id,
                    "is_complete": state_job.is_complete(),
                    "has_error": state_job.has_error(),
                }
            )

    return jsonify({"jobs": jobs})


@orchestrationsBP.route("/jobs/<job_id>/log", methods=["GET"])
def job_log(job_id) -> Response:
    """
    Endpoint for getting the most recent log generated by a job with job_id
    """
    project = Project.find()
    log_service = JobLoggingService(project)
    log = log_service.get_latest_log(job_id)

    finder = JobFinder(job_id)
    state_job = finder.latest(db.session)

    return jsonify(
        {
            "job_id": job_id,
            "log": log,
            "has_error": state_job.has_error() if state_job else False,
        }
    )


@orchestrationsBP.route("/run", methods=["POST"])
def run():
    project = Project.find()
    schedule_payload = request.get_json()
    job_id = run_elt(project, schedule_payload)

    return jsonify({"job_id": job_id}), 202


@orchestrationsBP.route("/<plugin_ref:plugin_ref>/configuration", methods=["GET"])
def get_plugin_configuration(plugin_ref) -> Response:
    """
    endpoint for getting a plugin's configuration
    """
    project = Project.find()
    settings = PluginSettingsService(project)
    config = flatten(
        settings.as_config(db.session, plugin_ref, redacted=True), reducer="dot"
    )

    return jsonify(
        {
            # freeze the keys because they are used for lookups
            "config": freeze_keys(config),
            "settings": settings.get_definition(plugin_ref).settings,
        }
    )


@orchestrationsBP.route("/<plugin_ref:plugin_ref>/configuration", methods=["PUT"])
def save_plugin_configuration(plugin_ref) -> Response:
    """
    endpoint for persisting a plugin configuration
    """
    project = Project.find()
    payload = request.get_json()

    settings = PluginSettingsService(project)
    for name, value in payload.items():
        # we want to prevent the edition of protected settings from the UI
        if settings.find_setting(plugin_ref, name).get("protected"):
            logging.warning("Cannot set a 'protected' configuration externally.")
            continue

        if value == "":
            settings.unset(db.session, plugin_ref, name)
        else:
            settings.set(db.session, plugin_ref, name, value)

    return jsonify(settings.as_config(db.session, plugin_ref, redacted=True))


@orchestrationsBP.route("/pipeline_schedules", methods=["GET"])
def get_pipeline_schedules():
    """
    endpoint for getting the pipeline schedules
    """
    project = Project.find()
    schedule_service = ScheduleService(project)
    schedules = [s._asdict() for s in schedule_service.schedules()]
    for schedule in schedules:
        finder = JobFinder(schedule["name"])
        state_job = finder.latest(db.session)
        schedule["has_error"] = state_job.has_error() if state_job else False
        schedule["is_running"] = state_job.is_running() if state_job else False
        schedule["job_id"] = state_job.job_id if state_job else None

    return jsonify(schedules)


@orchestrationsBP.route("/pipeline_schedules", methods=["POST"])
def save_pipeline_schedule() -> Response:
    """
    endpoint for persisting a pipeline schedule
    """
    incoming = request.get_json()
    # Airflow requires alphanumeric characters, dashes, dots and underscores exclusively
    name = slugify(incoming["name"])
    extractor = incoming["extractor"]
    loader = incoming["loader"]
    transform = incoming["transform"]
    interval = incoming["interval"]

    project = Project.find()
    schedule_service = ScheduleService(project)

    try:
        schedule = schedule_service.add(
            db.session, name, extractor, loader, transform, interval
        )
        return jsonify(schedule._asdict()), 201
    except ScheduleAlreadyExistsError as e:
        raise ScheduleAlreadyExistsError(e.schedule)



import logging
import sqlalchemy.types as types
import uuid
from datetime import datetime
from contextlib import contextmanager
from enum import Enum
from sqlalchemy import Column
from sqlalchemy.ext.mutable import MutableDict

from meltano.core.db import SystemModel
from meltano.core.error import Error
from meltano.core.sqlalchemy import JSONEncodedDict, IntFlag, GUID


class InconsistentStateError(Error):
    """
    Occur upon a wrong operation for the current state.
    """


class ImpossibleTransitionError(Error):
    """
    Occur upon a wrong transition.
    """


class State(Enum):
    IDLE = (0, ("RUNNING", "FAIL"))
    RUNNING = (1, ("SUCCESS", "FAIL"))
    SUCCESS = (2, ())
    FAIL = (3, ("RUNNING",))
    DEAD = (4, ())

    def transitions(self):
        return self.value[1]

    def __str__(self):
        return self.name


class Job(SystemModel):
    __tablename__ = "job"

    id = Column(types.Integer, primary_key=True)
    job_id = Column(types.String)
    run_id = Column(GUID, nullable=False, default=uuid.uuid4)
    state = Column(types.Enum(State, name="job_state"))
    started_at = Column(types.DateTime)
    ended_at = Column(types.DateTime)
    payload = Column(MutableDict.as_mutable(JSONEncodedDict))
    payload_flags = Column(IntFlag, default=0)

    def __init__(self, **kwargs):
        kwargs["state"] = kwargs.get("state", State.IDLE)
        kwargs["payload"] = kwargs.get("payload", {})
        super().__init__(**kwargs)

    def is_running(self):
        return self.state is State.RUNNING

    def has_error(self):
        return self.state is State.FAIL

    def is_complete(self):
        return self.state in [State.SUCCESS, State.FAIL]

    def can_transit(self, state: State) -> bool:
        if self.state is state:
            return True

        return state.name in self.state.transitions()

    def transit(self, state: State) -> (State, State):
        transition = (self.state, state)

        if not self.can_transit(state):
            raise ImpossibleTransitionError(transition)

        if self.state is state:
            return transition

        self.state = state

        return transition

    @contextmanager
    def run(self, session):
        try:
            self.start()
            self.save(session)

            yield

            self.success()
            self.save(session)
        except Exception as err:
            logging.error(err)
            self.fail(error=err)
            self.save(session)
            raise

    def start(self):
        self.started_at = datetime.utcnow()
        self.transit(State.RUNNING)

    def fail(self, error=None):
        self.ended_at = datetime.utcnow()
        self.transit(State.FAIL)
        if error:
            self.payload.update({"error": str(error)})

    def success(self):
        self.ended_at = datetime.utcnow()
        self.transit(State.SUCCESS)

    def __repr__(self):
        return "<Job(id='%s', job_id='%s', state='%s')>" % (
            self.id,
            self.job_id,
            self.state,
        )

    def save(self, session):
        session.add(self)
        session.commit()

        return self



"""Create meltano.core base tables

Revision ID: b4c05e463b53
Revises: 6f28844bcd3c
Create Date: 2019-07-23 16:05:29.073296

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.ext.mutable import MutableDict
from enum import Enum


# revision identifiers, used by Alembic.
revision = "b4c05e463b53"
down_revision = None
branch_labels = None
depends_on = None

# compat for types declared in `meltano.core.sqlalchemy`
# basically returns the `impl` type backing the field
JSONEncodedDict = sa.VARCHAR
IntFlag = sa.INTEGER


# from `src/meltano/core/job.py`
class State(Enum):
    IDLE = (0, ("RUNNING", "FAIL"))
    RUNNING = (1, ("SUCCESS", "FAIL"))
    SUCCESS = (2, ())
    FAIL = (3, ("RUNNING",))
    DEAD = (4, ())


def upgrade():
    op.create_table(
        "job",
        sa.Column("id", sa.Integer, primary_key=True),
        sa.Column("job_id", sa.String),
        sa.Column("state", sa.Enum(State, name="job_state")),
        sa.Column("started_at", sa.DateTime),
        sa.Column("ended_at", sa.DateTime),
        sa.Column("payload", MutableDict.as_mutable(JSONEncodedDict)),
        sa.Column("payload_flags", IntFlag, default=0),
    )

    op.create_table(
        "plugin_settings",
        sa.Column("label", sa.String(), nullable=True),
        sa.Column("description", sa.Text(), nullable=True),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("namespace", sa.String(), nullable=True),
        sa.Column("value", sa.PickleType(), nullable=True),
        sa.Column("enabled", sa.Boolean(), nullable=True),
        sa.PrimaryKeyConstraint("name", "namespace"),
    )


def downgrade():
    op.drop_table("job")
    op.drop_table("plugin_settings")



import asyncio
import logging
from io import StringIO

from . import Runner
from meltano.core.error import SubprocessError
from meltano.core.project import Project
from meltano.core.plugin import PluginType
from meltano.core.plugin.settings_service import PluginSettingsService
from meltano.core.connection_service import ConnectionService
from meltano.core.dbt_service import DbtService
from meltano.core.config_service import ConfigService
from meltano.core.db import project_engine
from meltano.core.elt_context import ELTContext


class DbtRunner(Runner):
    def __init__(self, elt_context: ELTContext, dbt_service: DbtService = None):
        self.context = elt_context
        self.dbt_service = dbt_service or DbtService(elt_context.project)
        self.connection_service = ConnectionService(elt_context)

    @property
    def project(self):
        return self.context.project

    def run(self, session, dry_run=False, models=None):
        # we should probably refactor this part to have an ELTContext object already
        # filled with the each plugins' configuration so we don't have to query
        # multiple times for the same data.

        settings_service = PluginSettingsService(self.project)
        try:
            load = self.connection_service.load_params()
            analyze = self.connection_service.analyze_params()
            env = {
                # inject the inferred 'schemas' from the ELTContext
                "MELTANO_LOAD_SCHEMA": load["schema"],
                "MELTANO_ANALYZE_SCHEMA": analyze["schema"],
                "DBT_TARGET": self.connection_service.dialect,
                # inject the extractor & loader configuration as ENV variables.
                # that means dbt will have access to all the configuration of
                # the extractor and loader
                **settings_service.as_env(session, self.context.extractor.ref),
                **settings_service.as_env(session, self.context.loader.ref),
            }
        except Exception as e:
            logging.debug("Could not inject environment to dbt.")
            logging.debug(f"Could not hydrate ENV from the EltContext: {str(e)}")
            raise e

        # Get an asyncio event loop and use it to run the dbt commands
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self.dbt_service.deps())

        if models is not None:
            models = models.replace("-", "_")

        if dry_run:
            loop.run_until_complete(self.dbt_service.compile(models, env=env))
        else:
            loop.run_until_complete(self.dbt_service.run(models, env=env))



import os
import json
from os.path import join
from pathlib import Path

from meltano.core.m5o.m5o_collection_parser import (
    M5oCollectionParser,
    M5oCollectionParserTypes,
)
from meltano.core.m5o.m5o_file_parser import MeltanoAnalysisFileParser
from meltano.core.project import Project
from meltano.core.utils import slugify
from .sql_helper import SqlHelper


class DashboardAlreadyExistsError(Exception):
    """Occurs when a dashboard already exists."""

    def __init__(self, dashboard):
        self.dashboard = dashboard


class DashboardsHelper:
    VERSION = "1.0.0"

    def get_dashboards(self):
        project = Project.find()
        dashboardsParser = M5oCollectionParser(
            project.analyze_dir("dashboards"), M5oCollectionParserTypes.Dashboard
        )

        return dashboardsParser.parse()

    def get_dashboard_reports_with_query_results(self, reports):
        sqlHelper = SqlHelper()

        for report in reports:
            m5oc = sqlHelper.get_m5oc_topic(report["namespace"], report["model"])
            design = m5oc.design(report["design"])
            loader = report["query_payload"]["loader"]

            sql_dict = sqlHelper.get_sql(design, report["query_payload"])
            outgoing_sql = sql_dict["sql"]
            aggregates = sql_dict["aggregates"]

            report["query_results"] = sqlHelper.get_query_results(loader, outgoing_sql)
            report["query_result_aggregates"] = aggregates

        return reports

    def get_dashboard(self, dashboard_id):
        dashboards = self.get_dashboards()
        target_dashboard = [
            dashboard for dashboard in dashboards if dashboard["id"] == dashboard_id
        ]

        return target_dashboard[0]

    def get_dashboard_by_name(self, name):
        dashboards = self.get_dashboards()
        dashboard = next(filter(lambda r: r["name"] == name, dashboards), None)

        return dashboard

    def save_dashboard(self, data):
        dashboard_name = data["name"]

        # guard if it already exists
        existing_dashboard = self.get_dashboard_by_name(dashboard_name)
        if existing_dashboard:
            raise DashboardAlreadyExistsError(existing_dashboard)

        project = Project.find()
        slug = slugify(dashboard_name)
        file_path = project.analyze_dir("dashboards", f"{slug}.dashboard.m5o")
        data = MeltanoAnalysisFileParser.fill_base_m5o_dict(file_path, slug, data)
        data["version"] = DashboardsHelper.VERSION
        data["description"] = data["description"] or ""
        data["report_ids"] = []

        with file_path.open("w") as f:
            json.dump(data, f)

        return data

    def add_report_to_dashboard(self, data):
        project = Project.find()
        dashboard = self.get_dashboard(data["dashboard_id"])

        if data["report_id"] not in dashboard["report_ids"]:
            dashboard["report_ids"].append(data["report_id"])
            file_path = project.analyze_dir(
                "dashboards", f"{dashboard['slug']}.dashboard.m5o"
            )
            with file_path.open("w") as f:
                json.dump(dashboard, f)

        return dashboard

    def remove_report_from_dashboard(self, data):
        project = Project.find()
        dashboard = self.get_dashboard(data["dashboard_id"])

        if data["report_id"] in dashboard["report_ids"]:
            dashboard["report_ids"].remove(data["report_id"])
            file_path = project.analyze_dir(
                "dashboards", f"{dashboard['slug']}.dashboard.m5o"
            )

            with file_path.open("w") as f:
                json.dump(dashboard, f)

        return dashboard
